# -*- coding: utf-8 -*-
# #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################

from .. import Tag
from ..render import HRenderer


import webbrowser,os,json
try:
    import uvicorn
    from starlette.applications import Starlette
    from starlette.responses import HTMLResponse
except ImportError:
    import sys
    print("You should install 'starlette' & 'uvicorn' for this htag runner")
    sys.exit(-1)


class BrowserStarletteWS:
    """ Simple ASync Web Server (with starlette) with WebSocket interactions with HTag.
        Open the rendering in a browser tab.
    """
    def __init__(self,tag:Tag):
        js = """
async function interact( o ) {
    ws.send( JSON.stringify(o) );
}

var ws = new WebSocket("ws://localhost:8000/ws");
ws.onopen = start;
ws.onmessage = function(e) {
    action( JSON.parse(e.data) );
};
"""

        self.renderer=HRenderer(tag, js, lambda: os._exit(0))


    def run(self):
        app = Starlette()

        @app.route('/')
        async def homepage(request):
            return HTMLResponse( str(self.renderer) )

        @app.websocket_route('/ws')
        async def websocket(websocket):
            await websocket.accept()
            while True:
                mesg = await websocket.receive_text()
                data = json.loads(mesg)
                actions = await self.renderer.interact(data["id"],data["method"],data["args"],data["kargs"])
                await websocket.send_text( json.dumps(actions) )
            await websocket.close()

        webbrowser.open_new_tab("http://127.0.0.1:8000")
        uvicorn.run(app)



# -*- coding: utf-8 -*-
# # #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################

from .. import Tag
from ..render import HRenderer

""" REAL WEB http,
- .exit() has no effect ;-)
- can handle multiple client (see SESID below)
- each client -> an instance is created (at post)
- session are destroyed after 5m of inactivity
"""

import os
import time
import asyncio

TIMEOUT = 5*60 # in seconds

try:
    import uvicorn
    from starlette.applications import Starlette
    from starlette.responses import HTMLResponse,JSONResponse
    from starlette.routing import Route
except ImportError:
    import sys
    print("You should install 'starlette' & 'uvicorn' for this htag runner")
    sys.exit(-1)


class WebHTTP:
    """ Simple ASync Web Server (with starlette) with HTTP interactions with htag.
        can handle multiple instances !!!
    """
    def __init__(self,callback_tag_creator):
        self.sessions={}
        self.callback_tag_creator = callback_tag_creator

    def _createRenderer(self):
        js = """
async function interact( o ) {
    action( await (await window.fetch("/"+SESID,{method:"POST", body:JSON.stringify(o)})).json() )
}

//======================================================= create a unique SESSION ID
var SESID = window.sessionStorage.getItem('SESID');
if(!SESID) {
    SESID = Math.random().toString(36).substring(2);
    window.sessionStorage.setItem('SESID',SESID);
}
console.log("SESSION ID:",SESID)
//=======================================================

window.addEventListener('DOMContentLoaded', start );
"""
        tag = self.callback_tag_creator()
        return HRenderer(tag, js ) # NO EXIT !!

    def _getRendererSession(self,ses_id):

        if ses_id in self.sessions: # reuse
            ses_renderer = self.sessions[ses_id]["renderer"]
        else: # create a new one
            ses_renderer = self._createRenderer()
            self.sessions[ses_id] = dict( lastaccess=None, renderer=ses_renderer )

        self.sessions[ses_id]["lastaccess"] = time.time()
        return self.sessions[ses_id]["renderer"]

    async def _purgeSessions(self):

        async def purge():
            while True:
                now=time.time()
                for ses_id,ses_info in list(self.sessions.items()):
                    if now - ses_info["lastaccess"] > TIMEOUT:
                        del self.sessions[ses_id]

                await asyncio.sleep(60) # every 60s

        asyncio.ensure_future( purge() )


    async def GET(self,request) -> HTMLResponse:
        # create a "empty shelf" (the renderer is destroyed after use)
        r=self._createRenderer()
        return HTMLResponse( str(r) )

    async def POST(self,request) -> JSONResponse:
        data=await request.json()

        # get renderer from session id
        ses_renderer = self._getRendererSession(request.path_params['sesid'])

        actions = await ses_renderer.interact(data["id"],data["method"],data["args"],data["kargs"])
        return JSONResponse(actions)

    def run(self, host="0.0.0.0", port=8000):   # wide, by default !!
        app = Starlette(debug=True, routes=[
            Route('/', self.GET, methods=["GET"]),
            Route('/{sesid}', self.POST, methods=["POST"]),
        ], on_startup=[self._purgeSessions])
        uvicorn.run(app, host=host, port=port)



# -*- coding: utf-8 -*-
# # #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################

from .. import Tag
from ..render import HRenderer

import webbrowser,os

try:
    import uvicorn
    from starlette.applications import Starlette
    from starlette.responses import HTMLResponse,JSONResponse
    from starlette.routing import Route
except ImportError:
    import sys
    print("You should install 'starlette' & 'uvicorn' for this htag runner")
    sys.exit(-1)


class BrowserStarletteHTTP:
    """ Simple ASync Web Server (with starlette) with HTTP interactions with htag.
        Open the rendering in a browser tab.
    """
    def __init__(self,tag:Tag):
        js = """
async function interact( o ) {
    action( await (await window.fetch("/",{method:"POST", body:JSON.stringify(o)})).json() )
}

window.addEventListener('DOMContentLoaded', start );
"""

        self.renderer=HRenderer(tag, js, lambda: os._exit(0))

    async def GET(self,request) -> HTMLResponse:
        return HTMLResponse( str(self.renderer) )

    async def POST(self,request) -> JSONResponse:
        data=await request.json()
        dico = await self.renderer.interact(data["id"],data["method"],data["args"],data["kargs"])
        return JSONResponse(dico)

    def run(self):
        app = Starlette(debug=True, routes=[
            Route('/', self.GET, methods=["GET"]),
            Route('/', self.POST, methods=["POST"]),
        ])
        webbrowser.open_new_tab("http://127.0.0.1:8000")
        uvicorn.run(app)



# -*- coding: utf-8 -*-
# # #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################
import json,asyncio,types,traceback

from . import __version__
from .tag import HTagException,H, Tag, TagBase, genJsInteraction

import logging
logger = logging.getLogger(__name__)

def fmtcaller(method,a,k) -> str:
    """ pretty format caller method( a:tuple, k:dict) -> str"""
    MAX=15
    def show(x):
        if isinstance(x,str):
            if len(x)>=MAX:
                x=x[:MAX-3]+"..."
            return json.dumps(x)
        elif isinstance(x,bytes):
            if len(x)>=MAX:
                x= x[:MAX-3]+b"..."
        return str(x)

    args = [show(i) for i in a]+[f"{k}={show(v)}" for k,v in k.items()]
    return f'{method}( %s )' % ", ".join(args)

class Stater:
    """ Class to save all states of 'tag' and children/embbeded Tags, before interaction!
        so, after interaction : .guess() can guess modified one
        (intelligent rendering)
    """
    def __init__(self, tag: Tag ):
        logger.debug("Stater.__init__(), save state before interactions ")
        self.tag=tag
        self._states={}

        def rec( childs ):
            for obj in childs:
                tag,childs= list(obj.items())[0]
                self._states[id(tag)] = tag._getStateImage()
                rec( childs )

        rec( [self.tag._getTree()])
        logger.debug("Stater.__init__(), save %s states", len(self._states))

    def guess(self,dontRedraw=None): # -> list<Tag>
        """ to be runned after interactions to guess whose are modifieds
            return modifieds tags
        """
        logger.debug("Stater.guess(), start guessing ....")
        modifieds=[]

        def rec( childs):
            for obj in childs:
                tag,childs= list(obj.items())[0]
                state_after = tag._getStateImage()
                state_before= self._states.get( id(tag) )
                if state_after != state_before:
                    if id(tag)==id(dontRedraw):
                        logger.debug("Don't redraw %s (but modified)", repr(dontRedraw))
                        # but perhaps childs ...
                        rec(childs)
                    else:
                        logger.debug("STATE BEFORE for %s = '%s'", repr(tag), state_before )
                        logger.debug("STATE AFTER  for %s = '%s'", repr(tag), state_after )
                        modifieds.append(tag)
                else:
                    # no need to see childs, coz the parent will redraw all (childs too)
                    rec(childs)

        rec( [self.tag._getTree()] )
        logger.debug("Stater.guess(), modifieds components : %s", [repr(i) for i in modifieds])
        return modifieds


class HRenderer:
    def __init__(self, tag: Tag, js:str, exit_callback=None):
        """ tag object will be setted as a 'body' tag ! """
        """ js should containt an interact method, and an initializer to call the start()"""
        if not isinstance(tag, Tag): raise HTagException("Can only render a Tag !")
        self.tag=tag
        self.tag.tag="body" # force first tag as body !!
        if exit_callback:
            # add an .exit() method on the tag !!!!!!!!!!!!!!!!!!!!!!
            self.tag.exit = exit_callback

        self._statics=[]
        ensureList=lambda x: x if isinstance(x,list) else [x]
        def rec( subclasses ):
            for c in subclasses:
                for i in ensureList(c.statics):
                    if isinstance(i,TagBase):
                        if isinstance(i,Tag):
                            logger.warning("Avoid to include dynamic Tag in statics! (it's converted)")
                        if i.md5 not in [j.md5 for j in self._statics]:
                            self._statics.append( i )
                rec(c.__subclasses__())

        rec(Tag.__subclasses__())

        logger.debug("HRenderer(), statics found : %s", [repr(i) for i in self._statics])

        js_base="""
function start() { %s }

function action( o ) {

    if(o.hasOwnProperty("update"))
        Object.keys(o["update"]).forEach(key => {
            document.getElementById( key ).outerHTML = o["update"][key];
        });
    if(o.hasOwnProperty("stream"))
        Object.keys(o["stream"]).forEach(key => {
            document.getElementById( key ).innerHTML += o["stream"][key];
        });

    if(o.hasOwnProperty("post")) eval( o["post"] );
    if(o.hasOwnProperty("next")) eval( o["next"] );
    if(o.hasOwnProperty("err"))  console.log( "ERROR", o["err"] );
}

%s
""" % (genJsInteraction(0), js)

        self._statics.append( H.script( js_base ))

        self._loop={}   # for savng generator, to keep them during GC

    async def interact(self,oid,method_name,args,kargs) -> dict:
        """ call the 'method_name' of pyobj 'id' with (args,kargs), return the pyobj/tag"""
        try:
            interaction_scripts=[] # interact js scripts
            next_js_call=None

            if oid==0:
                # start (not a real interaction, just produce the first rendering of the main tag (will be a body))
                logger.info("INTERACT INITIAL: %s",repr(self.tag))
                rep = self._mkReponse( [self.tag] )
                rep["update"]={0: list(rep["update"].values())[0]}  #INPERSONNATE (for first interact on id#0)
            else:
                def hookInteractScripts(obj,js):
                    interaction_scripts.append( obj._genIIFEScript(js) ) #IIFE !

                state = Stater(self.tag)
                setattr(Tag,"__call__", hookInteractScripts) # not great (with concurrencies)
                try:
                    norender=False

                    obj = Tag.find_tag(oid)
                    if obj:
                        # call the method
                        method=getattr(obj,method_name)

                        norender = hasattr(method,"_norender")

                        logger.info(f"INTERACT with METHOD {fmtcaller(method_name,args,kargs)} %s, of %s", "**NoRender**" if norender else "", repr(obj) )

                        if asyncio.iscoroutinefunction( method ):
                            r=await method(*args,**kargs)
                        else:
                            r=method(*args,**kargs)

                        if isinstance(r, types.AsyncGeneratorType) or isinstance(r, types.GeneratorType):
                            # save it, to avoid GC (and save the Tag/norender)
                            self._loop[id(r)] = dict(gen=r,tag=obj,norender= norender)

                    else:
                        obj = self._loop.get(oid,None)
                        if obj: # it's a existing generator
                            r, obj, norender = obj["gen"], obj["tag"], obj["norender"]
                            logger.info("INTERACT with GENERATOR %s(...) %s, of %s",r.__name__, "**NoRender**" if norender else "", repr(obj) )
                        else:
                            raise Exception(f"{oid} is not an existing Tag or generator (dead objects ?)?!")

                    ret=None
                    if isinstance(r, types.AsyncGeneratorType):
                        # it's a "async def yield"
                        try:
                            ret = await r.__anext__()
                            next_js_call = genJsInteraction(id(r))
                        except StopAsyncIteration:
                            del self._loop[ id(r) ]
                    elif isinstance(r, types.GeneratorType):
                        # it's a "def yield"
                        try:
                            ret = r.__next__()
                            next_js_call = genJsInteraction(id(r))
                        except StopIteration:
                            del self._loop[ id(r) ]
                    else:
                        # it's a simple method
                        assert r is None
                        ret=None

                    if norender:
                        rep= self._mkReponse(state.guess(obj) ) # avoid redraw obj
                    else:
                        rep= self._mkReponse(state.guess() )

                    if ret:
                        obj <= ret # add content in object
                        rep.setdefault("stream",{})[ id(obj) ] = str(ret)

                finally:
                    # clean the (fucking) situation ;-)
                    try: # to avoid del trouble in multi concurency
                        del Tag.__call__
                    except AttributeError:
                        pass

            # --> rep
            if interaction_scripts:
                # some js scripts was generated during interactions
                if "post" not in rep:
                    rep["post"]=""
                # add them
                rep["post"] += "\n".join(interaction_scripts)

            if next_js_call:
                # if there was generator, set the next js call !
                rep["next"]=next_js_call

        except Exception as e:
            print("ERROR",e)
            logger.error("Exception %s:\n%s", e, traceback.format_exc())
            rep={"err": str(e) }

        logger.info("RETURN --> %s",json.dumps(rep,indent=4))

        return rep

    def _mkReponse(self, tags) -> dict: # can't set a "script" key -> so named "post" (coz guy)
        rep={}

        scripts = []
        updates={}

        if tags:
            logger.debug("Force Tag rendering (for response): %s",[repr(i) for i in tags])
            for tag in tags:
                updates[id(tag)]=str(tag)

                if isinstance(tag,Tag):
                    scripts.extend( tag._getAllJs() )

        if updates:
            rep["update"]=updates

        if scripts:
            rep["post"]="\n".join( scripts )

        return rep

    def __str__(self) -> str:
        head=H.head()
        head <= H.meta(_charset="utf-8")
        head <= H.meta(_name="viewport",_content="width=device-width, initial-scale=1")
        head <= H.meta(_name="version",_content=f"HTag {__version__}")
        head <= H.title( self.title )
        for i in self._statics:
            head <= i._renderStatic()

        body=H.body( "Loading...", _id=0 ) # #INPERSONNATE (first interact on id #0)
        return "<!DOCTYPE html>"+str(H.html( [head,body] ))

    @property
    def title(self):
        return self.tag.__class__.__name__






# -*- coding: utf-8 -*-
# #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################
from .. import Tag
from ..render import HRenderer

"""
With this runner ... you could use guy.run(..), guy.runCef( ...) and guy.serve(...) methods (and auto reloader too)
(and deliver APK on android, see guy's docs)

LIMITATION : But multi client doesn't work using guy.serve().
(because HRenderer can manage only one instance at once)
(to handle multiple, you should modify this class to handle multiple HRenderer)
"""
try:
    import guy
except ImportError:
    import sys
    print("You should install 'guy' for this htag runner")
    sys.exit(-1)

class GuyApp(guy.Guy):
    """ Open the rendering in a GUY instance
    Interaction with buil-in guy websockets !
    """

    def __init__(self,tag: Tag):
        js = """
async function interact( o ) {
    action( await self.interact(o["id"], o["method"], o["args"], o["kargs"]) );
}

guy.init( start );
"""
        if type(tag.statics)==list:
            tag.statics.append(Tag.script(_src="guy.js"))   # add the js script "guy.js" (autogenerated by guy)
        else:
            tag.statics=[tag.statics, Tag.script(_src="guy.js")]

        self.renderer=HRenderer(tag, js, self.exit)

        super().__init__()

    async def interact(self,tagid:str,method:str,args,kargs): # guy binded method
        """ guy exposed method to interact"""
        return await self.renderer.interact(tagid,method,args,kargs)

    def render(self,path=None):
        """ Override guy.render(...) to produce a default rendering """
        return str(self.renderer)



# -*- coding: utf-8 -*-
# #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################

# mono instance
from .browserstarlettehttp import BrowserStarletteHTTP
from .browserstarlettews import BrowserStarletteWS
from .browserhttp import BrowserHTTP
from .pywebview import PyWebWiew
from .guyapp import GuyApp

# multi instance
from .webhttp import WebHTTP

__all__ = ["BrowserHTTP","PyWebWiew","GuyApp","BrowserStarletteHTTP","BrowserStarletteWS","WebHTTP"]


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['htag', 'htag.runners']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'htag',
    'version': '0.1.4',
    'description': 'GUI toolkit for building GUI toolkits (and create beautiful applications for mobile, web, and desktop from a single python3 codebase)',
    'long_description': '# HTag : "[H]TML Tag"\n\nThe descendant of [gtag](https://github.com/manatlan/gtag) ... but :\n\n * Not tied to [guy](https://github.com/manatlan/guy)\n * Able to be used in anything which can display html/js/css (pywebview, cefpython3, a browser, .... or guy)\n * A **lot lot lot better and simpler** (better abstractions/code/concepts)\n * "intelligent rendering" (redraw only component on state changes)\n\nIt\'s a GUI toolkit for building GUI toolkits ;-)\n\n[Changelog](changelog.md)\n\n[Available on pypi.org](https://pypi.org/project/htag/)\n\n**HTag** provides somes [`runners`](htag/runners) ootb. But they are just here, for the show. IRL: you should build your own, for your needs.\n\n## To have a look\n\nSee the [demo source code](https://github.com/manatlan/htag/blob/main/examples/demo.py)\n\nTo try it :\n\n    $ pip3 install htag pywebview\n    $ wget https://raw.githubusercontent.com/manatlan/htag/main/examples/demo.py\n    $ python3 demo.py\n\nThere will be docs in the future ;-)\n\n## ROADMAP to 1.0.0\n\n * rock solid (need more tests)\n * ~~top level api could change (Tag() -> create a Tag, Tag.mytag() -> create a TagBase ... can be a little bit ambiguous)~~\n * add a runner with WS with stdlib ? (not starlette!)\n * I don\'t really like the current way to generate js in interaction : need to found something more solid.\n * ~~the current way to initiate the statics is odd (only on real (embedded) Tag\'s) : should find a better way (static like gtag ?!)~~\n\n## In French\nSorte de FWK (orienté composants), où on code (coté backend) des objets python (en fait des objets "Tag"), qui ont une representation HTML avec des interactions, qui peuvent s\'executer dans tout ce qui est capable d\'afficher du html/js/css (pywebview, cefpython3, a browser, ....)\n\nLes "interactions" sont des actions émanants de la partie front vers le back, pour synchroniser l\'état de l\'objet (côté back), et retourner sa nouvelle représentation front.\nLa nature de ces interactions dépendent du `runner` utilisé (browser>ajax|websocket, guy>Websocket, pywebview>inproc)\n\nLe fwk permet surtout de fabriquer ses composants ... et il faudrait utiliser ces composants dans une appli.\n\nAutant le fwk permet des interactions avec js/front ... autant, il ne faudrait pas en faire dans les composants finaux : l\'idée, c\'est d\'abstraire toutes interactions js : de manière à ce que ça soit totallement transparent dans les composants finaux.\n\n',
    'author': 'manatlan',
    'author_email': 'manatlan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/manatlan/htag',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
# #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################

from .tag import H,Tag,HTagException

__version__ = "0.1.4" # updated from pypoetry.toml

__all__= ["Tag","HTagException"]

def _majVersion():
    """ called by build script (to update __version__ from toml)
        "python3 -c 'from htag import _majVersion; _majVersion()"
     """
    import re,tomlkit

    file='htag/__init__.py'
    content = open(file,'r+').read()
    v=tomlkit.parse(open('pyproject.toml').read())['tool']['poetry']['version']
    with open(file,'w+') as fid:
        fid.write( re.sub(r'__version__ = "[^"]+"','__version__ = "%s"'%v,content,1) )
    print("htag.__version__ UPDATED to '%s'"%v)



# -*- coding: utf-8 -*-
# # #############################################################################
# Copyright (C) 2022 manatlan manatlan[at]gmail(dot)com
#
# MIT licence
#
# https://github.com/manatlan/htag
# #############################################################################
import html,json,hashlib
import logging,types,asyncio
import weakref
logger = logging.getLogger(__name__)

class HTagException(Exception): pass

md5= lambda x: hashlib.md5(x.encode('utf-8')).hexdigest()

def stringify(obj):
    def my(obj):
        if isinstance(obj,bytes): # trick to convert b'' -> pure javascript
            return "<:<:%s:>:>" % obj.decode()
    return json.dumps(obj, default=my).replace('"<:<:',"").replace(':>:>"',"")

def genJsInteraction(id,method=None,args=None,kargs=None):
    interact=dict(id=id,method=method,args=args,kargs=kargs)
    return f"""interact( {stringify(interact)} );"""


class TagBase:
    """ This is a class helper to produce a "HTML TAG" """
    tag: str="div" # default one

    def __init__(self, content=None,**_attrs):
        self.set(content)

        self._attrs={}
        for k,v in _attrs.items():
            if not k.startswith("_"):
                raise HTagException(f"Can't set attributs without underscore ('{k}' should be '_{k}')") # for convention only ;-(
            else:
                self[ k[1:].replace("_","-") ] = v

        # compute a md5 (to indentify state for statics only now)
        # WARN : attrs or content change -> doesn't affect md5 !
        self.md5 = md5( str(self._attrs) + str(self._contents))

    def __le__(self, o):
        self.add(o)
        return o

    def clear(self):
        self._contents=[]

    def set(self,elt):
        """ a bit like .innerHTML setting (avoid clear+add)"""
        self.clear()
        self.add( elt)

    def add(self,elt):
        """ add an object or a list/tuple of objects """
        if elt is None:
            pass
        elif not isinstance(elt,str) and hasattr(elt,"__iter__"):
            for i in elt:
                self._contents.append(i)
        else:
            self._contents.append(elt)

    def __setitem__(self,attr,value):
        self._attrs[attr]=value
    def __getitem__(self,attr):
        return self._attrs.get(attr,None)

    def __str__(self):
        return self.__render__( self._attrs.items() , str )

    def _renderStatic(self):
        """ IT REMOVES @ID on tagbase too ;-( """
        mystr = lambda x: x._renderStatic() if isinstance(x,TagBase) else str(x)
        attrs = [(k,v) for k,v in self._attrs.items() if k != "id" ]
        return self.__render__(attrs,mystr)

    def __render__(self, attrs, mystr ):
        rattrs=[]
        for k,v in attrs:
            if v is not None:
                if isinstance(v,bool):
                    if v == True:
                        rattrs.append(k)
                else:
                    rattrs.append( '%s="%s"' % (k,html.escape( str(v) )) )
        return """<%(tag)s%(attrs)s>%(content)s</%(tag)s>""" % dict(
            tag=self.tag.replace("_","-"),
            attrs=" ".join([""]+rattrs) if rattrs else "",
            content="".join([mystr(i) for i in self._contents if i is not None]),
        )



    def _getStateImage(self) -> str: #TODO: could disapear (can make something more inteligent here!)
        """Return a str'image (state) of the object, for quick detection (see Stater())"""

        logger.debug("Force Tag rendering (for state image): %s",repr(self))
        str(self) # force a re-rendering (for builded lately)

        image=lambda x: "[%s]"%id(x) if isinstance(x,Tag) else str(x)
        return """%s%s:%s""" % (
            self.tag,
            self._attrs,
            [image(i) for i in self._contents],
        )

    def _getTree(self) -> dict:
        """ return a tree of TagBase childs """
        ll=[]

        for i in self._contents:
            if isinstance(i,TagBase):
                ll.append( i._getTree() )

        return {self:ll}

    def __repr__(self):
        return f"<{self.__class__.__name__}'{self.tag} {self._attrs.get('id')} (childs:{len(self._contents)})>"



class Binder:
    def __init__(self,btag_instance):
        self.__instance=btag_instance
    def __getattr__(self,method:str):
        m=hasattr(self.__instance,method) and getattr(self.__instance,method)
        if m and callable( m ):
            def _(*args,**kargs):
                return genJsInteraction(id(self.__instance),method,args,kargs)
            return _
        else:
            raise HTagException("Unknown method '%s' in '%s'"%(method,self.__instance.__class__.__name__))

    def __call__(self,callback,*args,**kargs):
        return Caller(self.__instance,callback,args,kargs)


class TagBaseCreator(type):
    def __getattr__(self,name:str):
        return type('TagBase%s' % name.capitalize(), (TagBase,), {**TagBase.__dict__,"tag":name})

class H(metaclass=TagBaseCreator): # Html
    def __init__(self):
        raise HTagException("no!")


class TagCreator(type):
    def __getattr__(self,name:str):
        if name == "H":
            return H
        else:
            return type('Tag%s' % name.capitalize(), (Tag,), {**Tag.__dict__,"tag":name})

class Caller:
    def __init__(self,instance, callback, args, kargs):
        self.instance = instance
        self.callback = callback
        self.args = args
        self.kargs = kargs

class Tag(TagBase,metaclass=TagCreator): # custom tag (to inherit)
    statics: list = [] # list of "Tag", imported at start

    __instances__ = weakref.WeakValueDictionary()

    js: str = None  # post script, useful for js/init when tag is rendered

    @classmethod
    def NoRender(cls,f):
        f._norender = True
        return f

    @classmethod
    def find_tag(cls, obj_id):
        return cls.__instances__.get(obj_id, None)

    def __init__(self, content=None,**_attrs):
        self._callbacks_={}
        attrs={}
        auto={}
        for k,v in _attrs.items():
            if k.startswith("_"):
                attrs[k]=v
            else:
                auto[k]=v

        self.__dict__.update(auto)

        if "_id" in attrs:
            raise HTagException("can't set the html attribut '_id'")
        else:
            attrs["_id"]=id(self)   # force an @id !
        TagBase.__init__(self,None, **attrs)
        self.set(content)
        Tag.__instances__[id(self)]=self

    # new mechanism (could replace self.bind.<m>())
    #-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

    async def __on__(self,eventjs,*a,**ka):
        logger.info(f"callback __on__ {eventjs} {a} {ka}")
        method = self._callbacks_[eventjs]

        if hasattr(method, '__self__') and method.__self__ == self:
            if asyncio.iscoroutinefunction( method ):
                r=await method(*a,**ka)
            else:
                r=method(*a,**ka)
        else:
            if asyncio.iscoroutinefunction( method ):
                r=await method(self,*a,**ka)
            else:
                r=method(self,*a,**ka)

        if isinstance(r, types.AsyncGeneratorType):
            async for i in r:
                yield i
        elif isinstance(r, types.GeneratorType):
            for i in r:
                yield i

    def __setitem__(self,attr,value):
        if type(value) in [types.FunctionType,types.MethodType]:
            key = "%s-%s" % (attr,id(self))
            self._callbacks_[key]=value
            newvalue = self.bind.__on__(key)
            logger.info("Assign event '%s' (simple callback) on %s" % (attr,repr(self)))
        elif isinstance(value,Caller):
            key = "%s-%s" % (attr,id(self))
            value.instance._callbacks_[key]=value.callback
            newvalue=value.instance.bind.__on__(key,*value.args,**value.kargs)
            logger.info("Assign event '%s' (handled) on %s" % (attr,repr(value.instance)))
        else:
            newvalue = value
        TagBase.__setitem__(self,attr,newvalue)
    #-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=


    @property
    def bind(self):
        """ to bind method ! and return its js repr"""
        return Binder(self)

    #to override
    def exit(self):
        print("exit() DOES NOTHING (should be overrided)")

    def _getAllJs(self) -> list:
        """ get a list of IIFE js declared script of this tag and its children"""
        ll=[]
        if self.js:
            logger.debug("Init Script (.js) found in %s --> '%s'",repr(self),self.js)
            ll.append( self._genIIFEScript( self.js ) ) #IIFE !

        def rec(childs):
            for i in childs:
                if isinstance(i,Tag):
                    ll.extend( i._getAllJs() )
                elif isinstance(i,TagBase):
                    rec(i._contents)

        rec(self._contents)

        return ll

    def _genIIFEScript(self,js:str) -> str:
        return f"(function(tag){{ {js}\n }})(document.getElementById('{id(self)}'));"

    def __str__(self):
        logger.debug("Tag.__str__() : render str for %s", repr(self))
        return TagBase.__str__(self)


